const path = require('path');
const fs = require('fs');
const http = require('http');
const querystring = require('querystring');
const crypto = require('crypto');

const mkdirp = require('mkdirp');
const express = require('express');
const bodyParser = require('body-parser');
const bodyParserJson = bodyParser.json();
const cookie = require('cookie');
const rollup = require('rollup');
const rollupPluginNodeResolve = require('rollup-plugin-node-resolve');
const rollupPluginCommonJs = require('rollup-plugin-commonjs');
const rollupPluginJson = require('rollup-plugin-json');
const eccrypto = require('eccrypto-sync');
const backendApiLib = require('./lib/backend-api');
const zeoHeaderPath = require.resolve('zeo-header');

const DEFAULTS = {
  host: '0.0.0.0',
  port: 6778,
  siteUrl: '/',
  crdsUrl: 'http://127.0.0.1:6777',
  dataDirectory: path.join(__dirname, 'data'),
  mining: true,
};

const _readFile = (p, opts) => new Promise((accept, reject) => {
  fs.readFile(p, opts, (err, data) => {
    if (!err) {
      accept(data);
    } else {
      reject(err);
    }
  });
});
const _requestRollup = p => rollup.rollup({
  entry: p,
  plugins: [
    rollupPluginNodeResolve({
      main: true,
      preferBuiltins: false,
    }),
    rollupPluginCommonJs(),
    rollupPluginJson(),
  ],
})
  .then(bundle => {
    const result = bundle.generate({
      moduleName: module,
      format: 'cjs',
      useStrict: false,
    });
    const {code} = result;
    const wrappedCode = '(function() {\n' + code + '\n})();\n';
    return wrappedCode;
  });

class Vrid {
  constructor({
    siteUrl = DEFAULTS.siteUrl,
    crdsUrl = DEFAULTS.crdsUrl,
    dataDirectory = DEFAULTS.dataDirectory,
    app = express(),
    mining = DEFAULTS.mining,
  } = {}) {
    this.siteUrl = siteUrl;
    this.crdsUrl = crdsUrl;
    this.dataDirectory = dataDirectory;
    this.app = app;
    this.mining = mining;
  }

  mountApp() {
    const {siteUrl, crdsUrl, dataDirectory, app, mining} = this;
    const tabs = ['avatar', 'keys'].concat(mining ? ['mining', 'monitor'] : []);
    const urls = ['/id', '/id/keys'].concat(mining ? ['/id/mining', '/id/monitor'] : []);

    const _ensureDataDirectory = () => new Promise((accept, reject) => {
      mkdirp(dataDirectory, err => {
        if (!err) {
          accept();
        } else {
          reject(err);
        }
      });
    });
    const backendApi = backendApiLib({crdsUrl});
    const _requestIndexJsRollup = () => _requestRollup(path.join(__dirname, 'lib', 'index.js'));

    return Promise.all([
      _ensureDataDirectory(),
      _readFile(path.join(__dirname, 'public', 'index.html'), 'utf8'),
      _requestIndexJsRollup(),
      _readFile(zeoHeaderPath, 'utf8'),
      _readFile(path.join(__dirname, 'lib', 'creaturejs.js'), 'utf8'),
      _readFile(path.join(__dirname, 'public', 'css', 'style.css'), 'utf8'),
      _readFile(path.join(path.dirname(zeoHeaderPath), '..', 'css', 'header.css'), 'utf8'),
    ])
      .then(([
        ensureDataDirectoryResult,
        indexHtml,
        indexJs,
        headerJs,
        creaturejsJs,
        styleCss,
        headerCss,
      ]) => {
        const cors = (req, res, next) => {
          res.set('Access-Control-Allow-Origin', req.get('Origin'));
          res.set('Access-Control-Allow-Headers', 'Content-Type');
          res.set('Access-Control-Allow-Credentials', true);

          next();
        };
        const cookieParser = (req, res, next) => {
          const cookieHeader = req.get('Cookie');

          if (cookieHeader) {
            req.cookie = cookie.parse(cookieHeader);
          }

          next();
        };
        const privateKeyParser = (req, res, next) => {
          const privateKey = req.cookie && req.cookie.privateKey;

          if (privateKey) {
            req.privateKey = privateKey;
          }

          next();
        };
        const ensurePrivateKeyDefault = (req, res, next) => {
          if (!req.privateKey) {
            const privateKey = crypto.randomBytes(32).toString('base64');
            req.privateKey = privateKey;

            _setCookie(res, 'privateKey', privateKey);
          }

          next();
        };
        const ensurePrivateKeyRespond = defaultJson => (req, res, next) => {
          if (req.privateKey) {
            next();
          } else {
            res.json(defaultJson);
          }
        };
        const _setCookie = (res, key, value) => {
          const valueString = typeof value === 'string' ? value : JSON.stringify(value);
          res.setHeader('Set-Cookie', cookie.serialize(key, valueString, {
            path: '/',
            httpOnly: false,
            maxAge: 60 * 60 * 24 * 7 * 52 * 10, // 10 years
          }));
        };

        const assetsIndexHtml = (req, res, next) => {
          const address = req.privateKey ? backendApi.getAddress(new Buffer(req.privateKey, 'base64')) : null;

          res.type('text/html');
          res.send(
            indexHtml
              .replace(/HOME/g, JSON.stringify(siteUrl))
              .replace(/TABS/g, JSON.stringify(tabs.join(',')))
              .replace(/URLS/g, JSON.stringify(urls.join(',')))
              .replace(/USERNAME/g, JSON.stringify(address || ''))
          );
        };
        const assetsIndexRequest = [cookieParser, privateKeyParser, ensurePrivateKeyDefault, assetsIndexHtml];
        app.get(/^\/id(?:\/keys)?(?:\?.*)?$/, assetsIndexRequest);
        if (mining) {
          app.get(/^\/id(?:\/mining|\/monitor)(?:\?.*)?$/, assetsIndexRequest);
        }
        app.get('/id/login', assetsIndexHtml);
        app.get('/id/assets/:asset', assetsIndexRequest);
        app.get('/id/createAsset', assetsIndexRequest);
        app.get('/id/buyAsset', assetsIndexRequest);
        app.get('/id/confirmPurchase', assetsIndexRequest);
        const publicImgStatic = express.static(path.join(__dirname, 'public', 'img'));
        app.use('/id/img', publicImgStatic);

        // for standalone
        app.get('/favicon.ico', express.static(path.join(__dirname, 'public')));
        app.get('/img/logo.svg', (req, res, next) => {
          req.url = '/logo.svg';
          publicImgStatic(req, res, next);
        });

        app.get('/id/js/index.js', (req, res, next) => {
          res.type('application/javascript');
          res.send(indexJs);
        });
        app.get('/id/js/header.js', (req, res, next) => {
          res.type('application/javascript');
          res.send(headerJs);
        });
        app.get('/id/js/creaturejs.js', (req, res, next) => {
          res.type('application/javascript');
          res.send(creaturejsJs);
        });
        app.get('/id/css/style.css', (req, res, next) => {
          res.type('text/css');
          res.send(styleCss);
        });
        app.get('/id/css/header.css', (req, res, next) => {
          res.type('text/css');
          res.send(headerCss);
        });

        app.options('/id/api/*', cors, (req, res, next) => {
          res.send();
        });
        app.get('/id/api/status', cors, (req, res, next) => {
          backendApi.requestStatus()
            .then(status => {
              res.json(status);
            })
            .catch(err => {
              res.status(err.status || 500);
              res.send(err.stack);
            });
        });
        app.get('/id/api/address', cors, cookieParser, privateKeyParser, ensurePrivateKeyDefault, (req, res, next) => {
          const {privateKey} = req;
          const privateKeyBuffer = new Buffer(privateKey, 'base64');
          const address = backendApi.getAddress(privateKeyBuffer);

          res.json({address});
        });
        const _isBasicAsset = asset => /^(?:[A-Z0-9]|(?!^)\-(?!$))+(\.(?:[A-Z0-9]|(?!^)\-(?!$))+)?$/.test(asset);
        const _getAssets = balances => Object.keys(balances)
          .filter(asset => _isBasicAsset(asset))
          .map(asset => ({
            asset: asset,
            quantity: balances[asset],
          }));
        app.get('/id/api/assets', cors, cookieParser, privateKeyParser, ensurePrivateKeyDefault, (req, res, next) => {
          const {privateKey} = req;
          const privateKeyBuffer = new Buffer(privateKey, 'base64');
          const address = backendApi.getAddress(privateKeyBuffer);

          backendApi.requestUnconfirmedBalances(address)
            .then(balances => {
              const assets = _getAssets(balances);
              res.json(assets);
            })
            .catch(err => {
              res.status(err.status || 500);
              res.send(err.stack);
            });
        });
        app.post('/id/api/send', cors, bodyParserJson, (req, res, next) => {
          const {body} = req;

          if (
            typeof body == 'object' && body &&
            typeof body.asset === 'string' &&
            typeof body.quantity === 'number' &&
            typeof body.srcAddress === 'string' &&
            typeof body.dstAddress === 'string' &&
            typeof body.privateKey === 'string'
          ) {
            const {asset, quantity, srcAddress, dstAddress, privateKey} = body;

            backendApi.requestCreateSend(asset, quantity, srcAddress, dstAddress, privateKey)
              .then(result => {
                res.json(result);
              })
              .catch(err => {
                res.status(err.status || 500);
                res.send(err.stack);
              });
          } else {
            res.status(400);
            res.send();
          }
        });
        app.post('/id/api/mint', cors, bodyParserJson, (req, res, next) => {
          const {body} = req;

          if (
            typeof body == 'object' && body &&
            typeof body.asset === 'string' &&
            typeof body.quantity === 'number' &&
            typeof body.privateKey === 'string'
          ) {
            const {asset, quantity, privateKey} = body;

            backendApi.requestCreateMint(asset, quantity, privateKey)
              .then(result => {
                res.json(result);
              })
              .catch(err => {
                res.status(err.status || 500);
                res.send(err.stack);
              });
          } else {
            res.status(400);
            res.send();
          }
        });
        app.post('/id/api/minter', cors, bodyParserJson, (req, res, next) => {
          const {body} = req;

          if (
            typeof body == 'object' && body &&
            typeof body.asset === 'string' &&
            typeof body.privateKey === 'string'
          ) {
            const {asset, privateKey} = body;

            backendApi.requestCreateMinter(asset, privateKey)
              .then(result => {
                res.json(result);
              })
              .catch(err => {
                res.status(err.status || 500);
                res.send(err.stack);
              });
          } else {
            res.status(400);
            res.send();
          }
        });
        app.post('/id/api/price', cors, bodyParserJson, (req, res, next) => {
          const {body} = req;

          if (
            typeof body == 'object' && body &&
            typeof body.asset === 'string' &&
            typeof body.price === 'number' &&
            typeof body.privateKey === 'string'
          ) {
            const {asset, price, privateKey} = body;

            backendApi.requestCreatePrice(asset, price, privateKey)
              .then(result => {
                res.json(result);
              })
              .catch(err => {
                res.status(err.status || 500);
                res.send(err.stack);
              });
          } else {
            res.status(400);
            res.send();
          }
        });
        app.post('/id/api/buy', cors, bodyParserJson, (req, res, next) => {
          const {body} = req;

          if (
            typeof body == 'object' && body &&
            typeof body.asset === 'string' &&
            typeof body.quantity === 'number' &&
            typeof body.price === 'number' &&
            typeof body.privateKey === 'string'
          ) {
            const {asset, price, privateKey} = body;

            backendApi.requestCreateBuy(asset, quantity, price, privateKey)
              .then(result => {
                res.json(result);
              })
              .catch(err => {
                res.status(err.status || 500);
                res.send(err.stack);
              });
          } else {
            res.status(400);
            res.send();
          }
        });
        app.get('/id/api/unconfirmedBalances/:address', cors, (req, res, next) => {
          const {address} = req.params;

          backendApi.requestUnconfirmedBalances(address)
            .then(result => {
              res.json(result);
            })
            .catch(err => {
              res.status(err.status || 500);
              res.send(err.stack);
            });
        });
        app.get('/id/api/unconfirmedPrice/:asset', cors, (req, res, next) => {
          const {asset} = req.params;

          backendApi.requestUnconfirmedPrice(asset)
            .then(result => {
              res.json(result);
            })
            .catch(err => {
              res.status(err.status || 500);
              res.send(err.stack);
            });
        });
        app.get('/id/api/blockcache', cors, (req, res, next) => {
          backendApi.requestBlockCache()
            .then(result => {
              res.json(result);
            })
            .catch(err => {
              res.status(err.status || 500);
              res.send(err.stack);
            });
        });
        app.get('/id/api/mempool', cors, (req, res, next) => {
          backendApi.requestMempool()
            .then(result => {
              res.json(result);
            })
            .catch(err => {
              res.status(err.status || 500);
              res.send(err.stack);
            });
        });
        app.post('/id/api/get', cors, bodyParserJson, (req, res, next) => {
          const {body} = req;

          if (
            typeof body == 'object' && body &&
            typeof body.address === 'string' &&
            typeof body.asset === 'string' &&
            typeof body.quantity === 'number'
          ) {
            const {address, asset, quantity} = body;

            backendApi.requestCreateGet(address, asset, quantity)
              .then(result => {
                res.json(result);
              })
              .catch(err => {
                res.status(err.status || 500);
                res.send(err.stack);
              });
          } else {
            res.status(400);
            res.send();
          }
        });
        app.post('/id/api/drop', cors, bodyParserJson, (req, res, next) => {
          const {body} = req;

          if (
            typeof body == 'object' && body &&
            typeof body.address === 'string' &&
            typeof body.asset === 'string' &&
            typeof body.quantity === 'number'
          ) {
            const {address, asset, quantity} = body;

            backendApi.requestCreateDrop(address, asset, quantity)
              .then(result => {
                res.json(result);
              })
              .catch(err => {
                res.status(err.status || 500);
                res.send(err.stack);
              });
          } else {
            res.status(400);
            res.send();
          }
        });
        app.get('/id/api/cookie/:key', cors, cookieParser, (req, res, next) => {
          const {data: dataString} = (req.cookie || {});

          if (dataString) {
            const data = _jsonParse(dataString);

            if (typeof data === 'object' && data !== null) {
              const {key} = req.params;
              const value = (key in data) ? data[key] : null;
              res.json(value);
            } else {
              res.json(null);
            }
          } else {
            res.json(null);
          }
        });
        app.post('/id/api/cookie/:key', cors, bodyParserJson, (req, res, next) => {
          const data = (() => {
            const {data: dataString} = (req.cookie || {});

            if (dataString) {
              const data = _jsonParse(dataString);

              if (typeof data === 'object' && data !== null) {
                return data;
              } else {
                return {};
              }
            } else {
              return {};
            }
          })();
          const {key} = req.params;
          const value = req.body;
          data[key] = value;

          _setCookie(res, 'data', data);

          res.send();
        });
        app.get('/id/api/data/:asset', cors, (req, res, next) => {
          const {asset} = req.params;

          const rs = fs.createReadStream(path.join(dataDirectory, asset + ':data'));
          rs.pipe(res);
          rs.on('error', err => {
            if (err.code === 'ENOENT') {
              res.status(404);
            } else {
              res.status(500);
            }

            res.json({
              error: err.stack,
            });
          });
        });
        app.put('/id/api/data/:asset', cors, cookieParser, privateKeyParser, (req, res, next) => {
          const {asset} = req.params;
          const address = req.privateKey ? backendApi.getAddress(new Buffer(req.privateKey, 'base64')) : null;

          if (address !== null) {
            backendApi.requestUnconfirmedMinter(asset)
              .then(minter => {
                if (minter === address) {
                  const ws = fs.createWriteStream(path.join(dataDirectory, asset + ':data'));
                  req.pipe(ws);
                  ws.on('error', err => {
                    res.status(500);
                    res.json({
                      error: err.stack,
                    });
                  });
                  ws.on('finish', () => {
                    res.send();
                  });
                } else {
                  res.status(401);
                  res.send();
                }
              })
              .catch(err => {
                res.status(500);
                res.json({
                  error: err.stack,
                });
              });
          } else {
            res.status(401);
            res.send();
          }
        });
        if (mining) {
          app.post('/id/api/startmine', cors, cookieParser, privateKeyParser, (req, res, next) => {
            const address = req.privateKey ? backendApi.getAddress(new Buffer(req.privateKey, 'base64')) : null;

            backendApi.requestMine(address)
              .then(() => {
                res.send();
              })
              .catch(err => {
                res.status(500);
                res.json({
                  error: err.stack,
                });
              });
          });
          app.post('/id/api/stopmine', cors, (req, res, next) => {
            backendApi.requestMine(null)
              .then(() => {
                res.send();
              })
              .catch(err => {
                res.status(500);
                res.json({
                  error: err.stack,
                });
              });
          });
          app.get('/id/api/minedBlocks', cors, (req, res, next) => {
            backendApi.requestMinedBlocks()
              .then(minedBlocks => {
                res.json(minedBlocks);
              })
              .catch(err => {
                res.status(500);
                res.json({
                  error: err.stack,
                });
              });
          });
        }

        return Promise.resolve();
      });
  }

  listen({
    host = DEFAULTS.host,
    port = DEFAULTS.port,
  } = {}) {
    return this.mountApp()
      .then(() => new Promise((accept, reject) => {
        http.createServer(this.app)
          .listen(port, host, err => {
            if (!err) {
              accept(`http://${host.replace(/^0\.0\.0\.0$/, '127.0.0.1')}:${port}/id`);
            } else {
              reject(err);
            }
          });
      }));
  }
}

const _jsonParse = s => {
  let error = null;
  let result;
  try {
    result = JSON.parse(s);
  } catch (err) {
    error = err;
  }
  if (!error) {
    return result;
  } else {
    return undefined;
  }
};

module.exports = opts => new Vrid(opts);

if (!module.parent) {
  const args = process.argv.slice(2);
  const _findArg = name => {
    for (let i = 0; i < args.length; i++) {
      const arg = args[i];
      const match = arg.match(new RegExp('^' + name + '=(.+)$'));
      if (match) {
        return match[1];
      }
    }
    return null;
  };
  const host = _findArg('host') || DEFAULTS.host;
  const port = parseInt(_findArg('port'), 10) || DEFAULTS.port;
  const crdsUrl = _findArg('crdsUrl') || DEFAULTS.crdsUrl;

  new Vrid({
    crdsUrl,
  }).listen({
    host,
    port,
  });
}
